from functools import wraps

from fennel.utils import get_aioredis


def aioclient(f):
    # Lazily initialise app.aioclient. This is needed because the app object is not
    # initialised in an async context (it's not awaited), but the functions in this
    # module will be used in an async context. We want to reuse a client with a
    # connection pool for efficiency.
    @wraps(f)
    async def wrapper(app, *args, **kwargs):
        if not app.aioclient:
            app.aioclient = await get_aioredis(app, app.settings.client_poolsize)
        return await f(app, *args, **kwargs)

    return wrapper



import json
from typing import Any, List

from fennel.client.aio.utils import aioclient
from fennel.exceptions import Timeout
from fennel.job import Job
from fennel.status import SENT


@aioclient
async def send(app, job, xid=None) -> List:
    """
    Send the job to the main task queue with the status SENT.
    """
    job = job.replace(status=SENT)
    tx = app.aioclient.multi_exec()
    tx.xadd(app.keys.queue, fields={"uuid": job.uuid}, message_id=xid or "*")
    tx.hmset_dict(app.keys.status(job), job.serialise())
    return await tx.execute()


@aioclient
async def result(app, job, timeout) -> Any:
    """
    Retrieve the result of executing the given `job`, blocking for `timeout` seconds
    before raising Timeout.
    """
    key = app.keys.result(job)
    value = await app.aioclient.brpoplpush(key, key, timeout)

    if value is None:
        raise Timeout
    return json.loads(value)


@aioclient
async def read_dead(app, batchsize=100):
    """
    Iterate over the dead-letter queue and return all job data.
    """
    return [x async for x in _iter_dlq(app, batchsize=batchsize)]


@aioclient
async def replay_dead(app, filter=lambda job: True, batchsize=100):
    """
    Iterate over the dead-letter queue and replay any jobs for which filter(job)
    evaluates to True. The default is to replay all jobs.
    """
    jobs = [x async for x in _iter_dlq(app, batchsize=batchsize)]
    results = []
    tx = app.aioclient.multi_exec()
    for xid, job in jobs:
        if filter(job):
            tx.hset(app.keys.status(job), "status", SENT)
            tx.xadd(app.keys.queue, {"uuid": job.uuid})
            tx.xdel(app.keys.dead, xid)
            results.append((xid, job))
    await tx.execute()
    return results


@aioclient
async def purge_dead(app, filter=lambda job: True, batchsize=100):
    """
    Iterate over the dead-letter queue and delete any jobs for which filter(job)
    evaluates to True. The default is to delete all jobs.
    """
    jobs = [x async for x in _iter_dlq(app, batchsize=batchsize)]
    results = []
    tx = app.aioclient.multi_exec()
    for xid, job in jobs:
        if filter(job):
            tx.delete(app.keys.status(job))
            tx.xdel(app.keys.dead, xid)
            results.append((xid, job))
    await tx.execute()
    return results


async def _iter_dlq(app, batchsize=100):
    """
    Iterate over the dead-letter queue and replay any jobs for which filter(job)
    evaluates to True. The default is to replay all jobs.
    """
    xid = "0"

    while True:
        messages = await app.aioclient.xread(
            [app.keys.dead], latest_ids=[xid], count=batchsize, timeout=None
        )
        if not messages:
            return

        assert all(stream == app.keys.dead for stream, _, _ in messages)
        xid = messages[-1][1]

        tx = app.aioclient.multi_exec()
        for _, xid, fields in messages:
            tx.hgetall(app.keys.status_prefix + f":{fields['uuid']}")
        response = await tx.execute()

        jobs = zip(
            [xid for _, xid, _ in messages],
            [Job.deserialise(fields) for fields in response],
        )

        for xid, job in jobs:
            yield (xid, job)



import datetime
import logging
import logging.config
import os
from io import StringIO
from typing import Any, Dict

import colorama
import structlog


def init_logging(level="debug", format="console"):
    level = level.upper()

    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": True,
            "formatters": {"standard": {"format": "%(message)s"}},
            "handlers": {
                "fennel.client": {
                    "level": level,
                    "formatter": "standard",
                    "class": "logging.StreamHandler",
                    "stream": "ext://sys.stderr",
                }
            },
            "loggers": {
                "fennel.client": {
                    "handlers": ["fennel.client"],
                    "level": level,
                    "propagate": False,
                },
                "fennel.worker": {
                    # To be used with a custom QueueHandler to prevent interleaving
                    # among the multiple processes.
                    "level": level,
                    "propagate": False,
                },
            },
        }
    )

    structlog.configure(
        logger_factory=logging.getLogger,
        processors=[
            _add_meta,
            structlog.stdlib.filter_by_level,
            structlog.processors.StackInfoRenderer(),
            structlog.processors.format_exc_info,
            structlog.processors.UnicodeDecoder(),
            structlog.processors.JSONRenderer() if format == "json" else _renderer,
        ],
    )


def _add_meta(logger: str, method_name: str, event_dict: Dict) -> Dict:
    event_dict = {
        "event": event_dict.pop("event"),
        "exc_info": event_dict.pop("exc_info", None),
        "extra": event_dict,
    }

    if not event_dict["extra"]:
        event_dict.pop("extra")

    event_dict["timestamp"] = datetime.datetime.utcnow().isoformat() + "Z"
    event_dict["level"] = method_name.upper()
    event_dict["pid"] = os.getpid()
    return event_dict


def _renderer(logger: str, method_name: str, event_dict: Dict) -> str:
    reset = colorama.Style.RESET_ALL
    bright = colorama.Style.BRIGHT
    dim = colorama.Style.DIM

    red = colorama.Fore.RED
    yellow = colorama.Fore.YELLOW
    blue = colorama.Fore.BLUE
    green = colorama.Fore.GREEN

    colours = {
        "CRITICAL": red + bright,
        "EXCEPTION": red + bright,
        "ERROR": red + bright,
        "WARNING": yellow + bright,
        "INFO": green + bright,
        "DEBUG": blue + bright,
    }

    s = StringIO()

    ts = event_dict.pop("timestamp", None)
    if ts:
        s.write(f"{reset}{dim}{str(ts)}{reset} ")

    level = event_dict.pop("level", None)
    if level:
        s.write(f"{colours[level]}{level:>9}{reset} ")

    event = event_dict.pop("event")
    s.write(f"{bright}{event:<17}{reset} ")

    pid = event_dict.pop("pid", None)
    if pid:
        s.write(f"{blue}pid{reset}={pid} ")

    def _format_extra(key: str, value: Any) -> None:
        s.write(f"{blue}{key}{reset}={value}{reset} ")

    extra = event_dict.pop("extra", {})

    job = extra.pop("job", None)
    if job:
        for k, v in job.items():
            _format_extra(k, v)

    for key, value in extra.items():
        _format_extra(key, value)

    stack = event_dict.pop("stack", None)
    exc = event_dict.pop("exception", None)

    if stack is not None:
        s.write("\n" + stack)
        if exc is not None:
            s.write("\n\n" + "=" * 88 + "\n")
    if exc is not None:
        s.write("\n" + exc)

    return s.getvalue()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fennel', 'fennel.client', 'fennel.client.aio', 'fennel.worker']

package_data = \
{'': ['*'], 'fennel.worker': ['lua/*']}

install_requires = \
['aioredis>=1.3.0,<2.0.0',
 'click>=7.0,<8.0',
 'colorama>=0.4.1,<0.5.0',
 'pydantic>=0.32.2,<0.33.0',
 'redis>=3.3,<4.0',
 'structlog>=20.0,<21.0',
 'uvloop>=0.14.0,<0.15.0']

entry_points = \
{'console_scripts': ['fennel = fennel.cli:cli']}

setup_kwargs = {
    'name': 'fennel',
    'version': '0.2.2',
    'description': 'A task queue for Python based on Redis Streams.',
    'long_description': None,
    'author': 'Matt Westcott',
    'author_email': 'm.westcott@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://fennel.readthedocs.io',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import json
from datetime import datetime

from fennel import models
from fennel.client.aio.utils import aioclient
from fennel.exceptions import JobNotFound
from fennel.job import Job


@aioclient
async def get_status(app, uuid) -> Job:
    resp = await app.aioclient.hget(app.keys.status_prefix + f":{uuid}", "status")
    if not resp:
        raise JobNotFound(uuid)
    return resp


@aioclient
async def get_job(app, uuid) -> Job:
    resp = await app.aioclient.hgetall(app.keys.status_prefix + f":{uuid}")
    if not resp:
        raise JobNotFound(uuid)
    return Job(**resp)


@aioclient
async def count_jobs(app):
    return len(
        [x async for x in app.aioclient.iscan(match=app.keys.status_prefix + ":*")]
    )


@aioclient
async def get_messages(app, key, n=100):
    messages = await app.aioclient.xrevrange(key, count=n)
    return [models.Message(id=id, uuid=fields["uuid"]) for id, fields in messages]


@aioclient
async def get_info(app, key):
    info = await app.aioclient.xinfo_stream(key)
    return models.Info(**{k.replace("-", "_"): v for k, v in info.items()})


@aioclient
async def get_groups(app, key):
    groups = await app.aioclient.xinfo_groups(key)
    return [
        models.Group(**{k.replace("-", "_"): v for k, v in d.items()}) for d in groups
    ]


@aioclient
async def get_stream(app, key, n=100):
    return models.Stream(
        key=key,
        info=await get_info(app, key),
        groups=await get_groups(app, key),
        messages=await get_messages(app, key, n=n),
    )


@aioclient
async def get_result(app, key):
    value = json.loads(await app.aioclient.lrange(key, 0, 1)[0])
    return models.Result(
        return_value=value["return_value"], exception=value["exception"]
    )


@aioclient
async def count_results(app):
    return len(
        [x async for x in app.aioclient.iscan(match=app.keys.result_prefix + ":*")]
    )


@aioclient
async def get_heartbeats(app):
    beats = await app.aioclient.hgetall(app.keys.heartbeats)
    return [
        models.Heartbeat(executor_id=k, timestamp=datetime.fromtimestamp(float(v)))
        for k, v in zip(beats.keys(), beats.values())
    ]


@aioclient
async def get_schedule(app):
    return [
        models.ScheduledJob(eta=datetime.fromtimestamp(int(v)), uuid=k)
        async for k, v in app.aioclient.izscan(app.keys.schedule)
    ]


@aioclient
async def get_state(app, messages=100):
    return models.State(
        queue=await get_stream(app, app.keys.queue, n=messages),
        dead=await get_stream(app, app.keys.dead, n=messages),
        heartbeats=await get_heartbeats(app),
        schedule=await get_schedule(app),
    )
